// Buffer class implementation
#include "pch.h"

#include "Buffer.h"
#include "util.h"

namespace nf::client::render {
    Buffer::Buffer(BufferType type, const VkDevice& device, const VkPhysicalDevice& physicalDevice, const VkCommandPool& commandPool, const VkQueue& queue, void* bufferData, size_t bufferSize)
        : VulkanResource(device)
        , m_handle()
        , m_memory()
    {
        VkBuffer stagingBuffer = nullptr;
        VkDeviceMemory stagingBufferMemory = nullptr;
        createBuffer(bufferSize, VK_BUFFER_USAGE_TRANSFER_SRC_BIT, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, physicalDevice, stagingBuffer, stagingBufferMemory);

        void* bufferPointer = nullptr;
        vkMapMemory(m_device, stagingBufferMemory, 0, bufferSize, 0, &bufferPointer);
        std::memcpy(bufferPointer, bufferData, bufferSize);
        vkUnmapMemory(m_device, stagingBufferMemory);

        VkBufferUsageFlags mainUsage = NULL;

        switch (type) {
            case BufferType::Vertex:
                mainUsage = VK_BUFFER_USAGE_VERTEX_BUFFER_BIT;
                break;

            case BufferType::Index:
                mainUsage = VK_BUFFER_USAGE_INDEX_BUFFER_BIT;
                break;
        }

        createBuffer(bufferSize, VK_BUFFER_USAGE_TRANSFER_DST_BIT | mainUsage, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT, physicalDevice, m_handle, m_memory);

        copyBuffer(stagingBuffer, m_handle, bufferSize, commandPool, queue);

        destroyBuffer(stagingBuffer, stagingBufferMemory);
    }

    const VkBuffer& Buffer::getHandle() const {
        return m_handle;
    }

    void Buffer::createBuffer(VkDeviceSize size, VkBufferUsageFlags usage, VkMemoryPropertyFlags memoryProperties, VkPhysicalDevice physicalDevice, VkBuffer& buffer, VkDeviceMemory& bufferMemory) {
        VkBufferCreateInfo bufferCI = { VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO };
        bufferCI.size = size;
        bufferCI.usage = usage;

        if (vkCreateBuffer(m_device, &bufferCI, nullptr, &buffer) != VK_SUCCESS)
            NFError("Could not create vertex buffer.");

        VkMemoryRequirements memoryRequirements = {};
        vkGetBufferMemoryRequirements(m_device, buffer, &memoryRequirements);

        VkMemoryAllocateInfo allocateInfo = { VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO };
        allocateInfo.allocationSize = memoryRequirements.size;
        allocateInfo.memoryTypeIndex = findMemoryType(physicalDevice, memoryRequirements.memoryTypeBits, memoryProperties);

        if (allocateInfo.memoryTypeIndex == -1)
            NFError("Could not find suitable memory type.");

        if (vkAllocateMemory(m_device, &allocateInfo, nullptr, &bufferMemory) != VK_SUCCESS)
            NFError("Could not allocate buffer memory.");

        vkBindBufferMemory(m_device, buffer, bufferMemory, 0);
    }

    uint32_t Buffer::findMemoryType(VkPhysicalDevice physicalDevice, uint32_t typeFilter, VkMemoryPropertyFlags properties) {
        VkPhysicalDeviceMemoryProperties memoryProperties = {};
        vkGetPhysicalDeviceMemoryProperties(physicalDevice, &memoryProperties);

        for (int i = 0; i < memoryProperties.memoryTypeCount; i++)
            if (typeFilter & (1 << i) && (memoryProperties.memoryTypes[i].propertyFlags & properties) == properties)
                return i;

        return -1;
    }

    void Buffer::copyBuffer(VkBuffer bufferSource, VkBuffer bufferDestination, VkDeviceSize size, VkCommandPool commandPool, VkQueue queue) {
        VkCommandBuffer commandBuffer = nullptr;
        VkCommandBufferAllocateInfo commandBufferAI = { VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO };
        commandBufferAI.commandPool = commandPool;
        commandBufferAI.commandBufferCount = 1;

        if (vkAllocateCommandBuffers(m_device, &commandBufferAI, &commandBuffer) != VK_SUCCESS)
            NFError("Could not create command buffer.");

        VkCommandBufferBeginInfo commandBufferBI = { VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO };
        commandBufferBI.flags = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT;

        if (vkBeginCommandBuffer(commandBuffer, &commandBufferBI) != VK_SUCCESS)
            NFError("Could not begin recording command buffer.");

        VkBufferCopy bufferCopyRegion = {};
        bufferCopyRegion.size = size;
        vkCmdCopyBuffer(commandBuffer, bufferSource, bufferDestination, 1, &bufferCopyRegion);

        if (vkEndCommandBuffer(commandBuffer) != VK_SUCCESS)
            NFError("Could not end command buffer.");

        VkSubmitInfo submitInfo = { VK_STRUCTURE_TYPE_SUBMIT_INFO };
        submitInfo.commandBufferCount = 1;
        submitInfo.pCommandBuffers = &commandBuffer;

        if (vkQueueSubmit(queue, 1, &submitInfo, nullptr) != VK_SUCCESS)
            NFError("Could not submit to Vulkan queue.");

        vkQueueWaitIdle(queue);

        vkFreeCommandBuffers(m_device, commandPool, 1, &commandBuffer);
    }

    void Buffer::destroyBuffer(VkBuffer buffer, VkDeviceMemory bufferMemory) {
        vkFreeMemory(m_device, bufferMemory, nullptr);
        vkDestroyBuffer(m_device, buffer, nullptr);
    }

    Buffer::~Buffer() {
        destroyBuffer(m_handle, m_memory);
    }
}
